import hashlib
import os
import regex as re
from nltk.tokenize import WordPunctTokenizer

class SpaceTokenizer:
    """
    Tokenize by splitting the text on space characters.
    """
    def __init__(self):
        pass

    def tokenize(self, text: str):
        return text.split(' ')


class StopWords:
    """
    Stop word remover using existing lists.
    """
    def __init__(self, custom_stop_words=[], ignore_numbers=False, ignore_punctuation=True, tokenizer=None):
        self.stop_words = self._get_stop_words(custom_stop_words)
        self.ignore_numbers = ignore_numbers
        self.ignore_punctuation = ignore_punctuation
        self.tokenizer = self._get_tokenizer(tokenizer)

    @staticmethod
    def _get_stop_words(custom_stop_words):
        stop_words = {}
        stop_word_dir = os.path.join(os.path.dirname(__file__), 'stop_words')
        for f in os.listdir(stop_word_dir):
            with open('{0}/{1}'.format(stop_word_dir, f), encoding="utf8") as fh:
                for stop_word in fh.read().strip().split('\n'):
                    stop_words[stop_word] = True

        for custom_stop_word in custom_stop_words:
            stop_words[custom_stop_word] = True

        return stop_words


    def _get_tokenizer(self, tokenizer):
        if not tokenizer:
            # If tokenizer is not defined, use simple space tokenizer
            return SpaceTokenizer()
        return tokenizer


    def _is_numerical(self, word):
        if self.ignore_numbers and re.fullmatch(r"\d+", word):
            return True
        return False


    def _is_punctuation(self, word):
        if self.ignore_punctuation and re.fullmatch(r"\W+", word):
            return True
        return False


    def remove(self, text):
        if isinstance(text, str):
            return ' '.join([lemma for lemma in self.tokenizer.tokenize(text) if lemma not in self.stop_words and not self._is_numerical(lemma) and not self._is_punctuation(lemma)])

        elif isinstance(text, list):
            return [lemma for lemma in text if lemma not in self.stop_words and not self._is_numerical(lemma) and not self._is_punctuation(lemma)]

        else:
            return None


class TextProcessor:
    """
    Processor for processing texts prior to modelling
    """

    def __init__(self, lemmatizer=None, phraser=None, tokenizer=None, remove_stop_words=True, sentences=False, 
                 words_as_list=False, custom_stop_words=[], input_texts=None, ignore_numbers=False, ignore_punctuation=False, text_hashing=False):
        # processing resources
        self.lemmatizer = lemmatizer
        self.phraser = phraser
        self.tokenizer = tokenizer
        self.word_punct_tokenizer = WordPunctTokenizer()
        self.space_tokenizer = SpaceTokenizer()

        self.stop_words = StopWords(custom_stop_words=custom_stop_words, ignore_numbers=ignore_numbers, ignore_punctuation=ignore_punctuation)
        # processing options
        self.remove_stop_words = remove_stop_words
        self.sentences = sentences
        self.words_as_list = words_as_list
        self.text_hashing = text_hashing
        # this is for when processor is used as an iterator (e.g. for gensim training)
        self.input_texts = input_texts



    def __iter__(self):
        if self.input_texts:
            for input_text in self.input_texts:
                processed = self.process(input_text)
                if processed:
                    yield processed[0]


    def _tokenize_and_lemmatize(self, text):
        """
        Tokenizes and lemmatizes text using MLP if asked.
        :param: str text: Input text to be tokenized & lemmatized.
        :return: Tokenized & lemmatized text.
        """
        # MLP lemmatizer tokenizes text anyway, so we don't need to do it again
        if self.lemmatizer:
            return self.lemmatizer.lemmatize(text)
        # tokenize without lemmatizing
        if self.tokenizer:
            return self.tokenizer.process(text)["text"]["text"]
        # return input text if nothing done
        return text
    

    def _hash_tokens(self, tokens):
        """
        Replaces tokens with MD5 hashes.
        """
        if self.text_hashing:
            return [hashlib.md5(token.encode()).hexdigest() for token in tokens]
        else:
            return tokens


    def process(self, input_text):
        if isinstance(input_text, str):
            stripped_text = input_text.strip()
            if self.sentences:
                list_of_texts = stripped_text.split('\n')
            else:
                list_of_texts = [stripped_text]
        else:
            # whetever obscure was found, output is as string
            list_of_texts = [str(input_text)]
        out = []
        for text in list_of_texts:
            if text:
                # make sure it is a string
                text = str(text)
                # tokenize & lemmatize if asked
                text = self._tokenize_and_lemmatize(text)
                # lower & strip
                text = text.lower().strip()
                # apply word punct tokenizer
                tokens = self.word_punct_tokenizer.tokenize(text)
                # convert string to list of tokens
                #tokens = self.space_tokenizer.tokenize(text)
                # remove stop words
                if self.remove_stop_words:
                    tokens = self.stop_words.remove(tokens)
                # use phraser
                if self.phraser:
                    tokens = self.phraser.phrase(tokens)
                # remove empty tokens
                tokens = [token for token in tokens if token]
                # prepare output
                if not self.words_as_list:
                    tokens = [token.replace(' ', '_') for token in tokens]
                    tokens = self._hash_tokens(tokens)
                    out.append(' '.join(tokens))
                else:
                    tokens = self._hash_tokens(tokens)
                    out.append(tokens)

        # return list of strings
        if self.sentences or self.words_as_list:
            return out
        # return first value as string
        elif out:
            return out[0]
        else:
            return ""



from gensim.models.phrases import Phraser as GSPhraser
from gensim.models import word2vec, fasttext, KeyedVectors
from gensim.models.phrases import Phrases
import joblib
import json

from texta_tools.text_processor import TextProcessor
from . import exceptions


class Phraser:
    """
    Wrapper of Gensim Phraser.
    """
    def __init__(self):
        self._phraser = None

    def train(self, sentences):
        """
        Trains phraser model using input sentences.
        """
        phrase_model = Phrases(sentences)
        phraser = GSPhraser(phrase_model)
        self._phraser = phraser

    def phrase(self, text):
        """
        Phrases input text.
        """
        if self._phraser:
            if isinstance(text, str):
                text = text.split(' ')
                return ' '.join(self._phraser[text])
            else:
                return self._phraser[text]
        else:
            return text


class Embedding:
    """
    Embedding Abstraction to work with both Word2Vec & FastText.
    """
    def __init__(self, description="My Embedding", workers=1, min_freq=5, num_dimensions=100, window=5, num_epochs=5,
                 text_processor=TextProcessor(sentences=True, remove_stop_words=True, words_as_list=True)):
        self.model = None
        self.phraser = None
        self.description = description
        # params
        self.workers = workers
        self.min_freq = min_freq
        self.window = window
        self.num_epochs = num_epochs
        self.num_dimensions = num_dimensions
        self.text_processor = text_processor

    def _train(self, texts, selected_model, use_phraser):
        """
        Trains Embedding.
        This needs to be called out in train() method of the model implementation (e.g. W2V).
        :param: list texts: List of texts or an iterator (e.g. Elasticsearcher).
        """
        if not texts:
            raise exceptions.InvalidInputError("No training texts provided.")
        # add texts to text processor so we can use it as an iterator
        self.text_processor.input_texts = texts
        # build phraser if asked
        if use_phraser == True:
            phraser = Phraser()
            phraser.train(self.text_processor)
            # set phraser
            self.phraser = phraser
            # update phraser in text processor
            self.text_processor.phraser = phraser

        # word2vec model
        model = selected_model(
            self.text_processor,
            min_count=self.min_freq,
            size=self.num_dimensions,
            iter=int(self.num_epochs),
            window=self.window,
            workers=self.workers
        )
        self.model = model
        return True

    def save(self, file_path):
        """
        Saves embedding with phraser to disk.
        """
        to_dump = {"phraser": self.phraser, "embedding": self.model}
        joblib.dump(to_dump, file_path)
        return True

    def load(self, file_path):
        """
        Loads embedding with phraser from disk.
        """
        to_load = joblib.load(file_path)
        self.model = to_load["embedding"]
        self.phraser = to_load["phraser"]
        return True

    def load_django(self, embedding_object):
        """
        Loads embedding in Django.
        """
        file_path = embedding_object.embedding_model.path
        to_load = joblib.load(file_path)
        self.model = to_load["embedding"]
        self.phraser = to_load["phraser"]
        self.description = embedding_object.description
        return True

    def get_vector(self, word):
        """
        Returns vector for given embedding entry.
        """
        if word not in self.model.wv.index2word:
            raise exceptions.OutOfVocabError("Word or phrase not in vocabulary.")
        return self.model.wv.__getitem__(word)

    def get_vocabulary(self):
        """
        Returns embedding vocabulary from KeyedVectors.
        """
        return self.model.wv.index2word


    @staticmethod
    def _format_str(string):
        return str(string).replace(' ', '_')


    def get_similar(self, positives_used, negatives_used=list(), positives_unused=list(), negatives_unused=list(), n=20):
        """
        Find similar words & phraser for input list of strings.
        """
        # check if all inputs are strings
        for input_list in (positives_used, negatives_used, positives_unused, negatives_unused):
            if not isinstance(input_list, list):
                raise exceptions.InvalidInputError("Input must be list!")
        # reformat positives & negatives
        # filter out words not present in the embedding vocabulary
        positives = [self._format_str(positive) for positive in positives_used if self._format_str(positive) in self.model.wv.vocab]
        negatives = [self._format_str(negative) for negative in negatives_used if self._format_str(negative) in self.model.wv.vocab]
        # create set of the tokens we don't want to see in the output
        neg_pos_combined = negatives_used + positives_unused + negatives_unused
        not_suggestions = [self._format_str(not_suggestion) for not_suggestion in neg_pos_combined if self._format_str(not_suggestion) in self.model.wv.vocab]
        not_suggestions = set(not_suggestions)
        # return logic
        if positives:
            similar_items = self.model.wv.most_similar(positive=positives, negative=negatives, topn=n+len(not_suggestions))
            similar_items_topn = list()
            for i,s in enumerate(similar_items):
                if s[0] not in not_suggestions:
                    similar_items_topn += [{'phrase': s[0].replace('_', ' '), 'score': s[1], 'model': self.description}]
                if i == n:
                    break
            return similar_items_topn
        else:
            return []


class FastTextEmbedding(Embedding):
    """
    This is an actual embedding class you want to use!
    """
    def train(self, texts, use_phraser=True):
        """
        Trains FastText embedding.
        :param: list texts: List of texts or an iterator (e.g. Elasticsearcher).
        """
        return self._train(texts, fasttext.FastText, use_phraser=use_phraser)


class W2VEmbedding(Embedding):
    """
    This is an actual embedding class you want to use!
    """
    def train(self, texts, use_phraser=True):
        """
        Trains Word2Vec embedding.
        :param: list texts: List of texts or an iterator (e.g. Elasticsearcher).
        """
        return self._train(texts, word2vec.Word2Vec, use_phraser=use_phraser)



import regex as re
from typing import List, Dict, Union, Tuple
from . import exceptions

SPLIT_OPTIONS = [
    "DOUBLE_NEWLINE",
    "NEWLINE",
    "WORD_LIMIT",
    "CHAR_LIMIT",
    "CUSTOM"
]

# what if doc consists of one looooong string without any whitespaces?

class TextSplitter:
    """ Split text into subsections based on given character/word limit.

    Parameters
    ----------
    split_by : str
        Method used for splitting the text.
    split_pattern : str
        Custom character/string_pattern for splitting the text.
        Has effect only if split_by = 'CUSTOM'.
    strip_whitespaces : bool
        If enabled, whitespaces (newlines etc.) are stripped from each page.
    keep_whitespace_formatting : bool
        If enabled, whitespace characters in the middle of each text chunk are left untouched,
        otherwise they will be removed, which will impacting the format.

    Attributes
    ------------
    split_by : str
        Stores parameter `split_by`.
    split_pattern : str
        Stores parameter `split_pattern`.
    strip_whitespaces : bool
        Stores parameter `strip_whitespaces`.
    keep_whitespace_formatting : bool
        Stores parameter `keep_whitespace_formatting`
    default_word_limit : int
        Default max number of words per one chunk.
    default_char_limit : int
        Default max number of characters per one chunk.
    """
    def __init__(self, split_by: str = "CHAR_LIMIT", split_pattern: str = "", strip_whitespaces: bool = True, keep_whitespace_formatting: bool = True):
        self.split_by = self._validate_split_option(split_by)
        self.split_pattern = split_pattern
        self.strip_whitespaces = strip_whitespaces
        self.keep_whitespace_formatting = keep_whitespace_formatting
        self.default_word_limit = 500
        self.default_char_limit = 3000

    def _validate_split_option(self, split_by: str):
        if split_by not in SPLIT_OPTIONS:
            raise exceptions.InvalidSplitMethodError(f"""Invalid argument for param `split_by`: `{split_by}`.
                Make sure to use one of the following: {SPLIT_OPTIONS}""")
        return split_by

    def get_split_options(self) -> List[str]:
        """ Retrieve list of available options for splitting the text.
        """
        return SPLIT_OPTIONS

    def _preprocess_text(self, text: str) -> str:
        """ Apply enabled text processing methods.

        Parameters
        ------------
        text : str
            Text on which to apply preprocessors.
        """
        if not self.keep_whitespace_formatting:
            # Remove extra whitespaces
            text = re.sub(r"\s+", " ", text)
        return text

    def _split_by_pattern(self, text: str, split_pattern: str) -> List[Dict[str, Union[str, int]]]:
        """ Split by string pattern specified with param `split_pattern`.

        Parameters
        ----------
        text : str
            Text to split.
        split_pattern : str
            Pattern used for splitting.
        """
        if self.strip_whitespaces:
            splits = [t.strip() for t in text.split(split_pattern) if t.strip()]
        else:
            splits = [t for t in text.split(split_pattern) if t.strip()]

        return [{"text": t, "page": i+1} for i, t in enumerate(splits)]

    def _split_by_word_limit(self, text: str, max_limit: int) -> List[Dict[str, Union[str, int]]]:
        """ Split into chunks of length `max_limit` (each chunk consists of `max_limit` words).

        Parameters
        ----------
        text : str
            Text to split.
        max_limit: int
            Maximum number of words per one chunk.
        """
        chunks = []
        words = [t for t in text.split()]

        n_chunks = int(len(words)/max_limit)

        if self.keep_whitespace_formatting:
            separators = re.findall(r"\s+", text)

        w_index = 0
        for i in range(n_chunks+1):
            slice_start = i*max_limit
            slice_end = i*max_limit + max_limit

            chunk_words = words[slice_start:slice_end]

            if self.keep_whitespace_formatting:
                chunk = ""
                for word in chunk_words:
                    chunk+=word
                    if w_index < len(separators):
                        chunk+=separators[w_index]
                        w_index+=1

            else:
                chunk = " ".join(chunk_words)

            # Remove trailing whitespaces
            if self.strip_whitespaces:
                chunk = chunk.strip()

            if chunk:
                chunks.append(chunk)

        return [{"text": t, "page": i+1} for i, t in enumerate(chunks)]

    def _split_by_char_limit(self, text: str, max_limit: int) -> List[Dict[str, Union[str, int]]]:
        """ Split into chunks of length `max_limit` (each chunk consists of `max_limit` characters).

        Parameters
        ----------
        text : str
            Text to split.
        max_limit: int
            Maximum number of characters per one chunk.
        """
        chunks = []
        n_chunks = int(len(text)/max_limit)

        offset = 0
        for i in range(n_chunks+1):
            slice_start = i*max_limit + offset
            slice_end = i*max_limit + max_limit + offset

            chunk = text[slice_start:slice_end]

            # Add non-whitespace characters from the begging of next chunk
            # to avoid splitting words:
            for c in text[slice_end:]:
                if not re.match(r"\s", c):
                    chunk+=c
                    offset+=1
                else:
                    break

            # Remove trailing whitespaces
            if self.strip_whitespaces:
                chunk = chunk.strip()

            if chunk:
                chunks.append(chunk)

        return [{"text": t, "page": i+1} for i, t in enumerate(chunks)]

    def split(self, text: str, max_limit: int = None) -> List[Dict[str, Union[str, int]]]:
        """ Split text into smaller chunks.

        Parameters
        ----------
        text : str
            Text to split.
        max_limit : int
            Max number of words or characters per each chunk. Has effect only if `split_by == 'WORD_LIMIT'` or
            `split_by == 'CHAR_LIMIT'`.

        Returns
        --------
        split_result: List[Dict[str, Union[str, int]]]
            List of dicts with text chunks and corresponding page numbers, e.g:
            [{'text': chunk_1, 'page': 1}, {'text': chunk_2, 'page': 2}]

        """
        # Apply enabled preprocessors
        text = self._preprocess_text(text)

        if self.split_by == "DOUBLE_NEWLINE":
            split_result = self._split_by_pattern(text, "\n\n")

        elif self.split_by == "NEWLINE":
            split_result = self._split_by_pattern(text, "\n")

        elif self.split_by == "WORD_LIMIT":
            if not max_limit:
                max_limit = self.default_word_limit
            split_result = self._split_by_word_limit(text, max_limit)

        elif self.split_by == "CHAR_LIMIT":
            if not max_limit:
                max_limit = self.default_char_limit
            split_result = self._split_by_char_limit(text, max_limit)

        elif self.split_by == "CUSTOM":
            split_result = self._split_by_pattern(text, self.split_pattern)

        return split_result



class MLPNotAvailableError(Exception):
    """Raised when MLP is not available."""
    pass

class MLPFailedError(Exception):
    """Raised when MLP processing fails."""
    pass

class InvalidInputError(Exception):
    """Raised when something incorrect given to trainers."""
    pass

class OutOfVocabError(Exception):
    """Raised when word not present in the vocabulary of embedding."""
    pass

class InvalidSplitMethodError(Exception):
    """Raised when TextSplitter is initiated with invalid splitting option."""
    pass



from urllib.parse import urljoin
import requests
import os

from . import exceptions


# set defaults for MLP API adapter
MLP_HOST = os.getenv("TEXTA_TAGGER_MLP_URL", "http://mlp-dev.texta.ee:5000")
MLP_MAJOR_VERSION = int(os.getenv("TEXTA_TAGGER_MLP_MAJOR_VERSION", "2"))


def check_mlp_connection(func):
    """
    Wrapper function for checking MLP connection prior to requests.
    """
    def func_wrapper(*args, **kwargs):
        MLP_URL = args[0].mlp_host
        try:
            response = requests.get(MLP_URL, timeout=3)
            if not response.ok: raise exceptions.MLPNotAvailableError()
            return func(*args, **kwargs)
        except (requests.exceptions.ConnectTimeout, requests.exceptions.ConnectionError) as e:
            raise exceptions.MLPNotAvailableError()
    return func_wrapper


class MLP2Analyzer:

    def __init__(self, MLP_HOST=MLP_HOST):
        self.mlp_host = MLP_HOST
        self.mlp_url = urljoin(MLP_HOST, "mlp")

    @check_mlp_connection
    def process(self, text):
        if not text or len(text) < 2:
            return {"text": {"text": text}, "texta_facts": []}
        response = requests.post(self.mlp_url, data=text.encode())
        if response.status_code != 200:
            raise exceptions.MLPFailedError("MLP service sent non-200 response. Please check MLP url and input.")
        response_json = response.json()
        return response_json

    @check_mlp_connection
    def lemmatize(self, text):
        if not text or len(text) < 2:
            return text
        response = requests.post(self.mlp_url, data=text.encode())
        if response.status_code != 200:
            raise exceptions.MLPFailedError(f"MLP service sent non-200 response: {response.text}")
        response_json = response.json()
        lemmas = response_json["text"]["lemmas"]
        return lemmas


class MLP3Analyzer:

    def __init__(self, MLP_HOST=MLP_HOST):
        self.mlp_host = MLP_HOST
        self.mlp_url = urljoin(MLP_HOST, "mlp/text")

    @check_mlp_connection
    def process(self, text):
        if not text:
            return {"text": {"text": text}, "texta_facts": []}
        payload = {"text": text.encode(), "analyzers": []}
        response = requests.post(self.mlp_url, data=payload)
        if response.status_code != 200:
            raise exceptions.MLPFailedError("MLP service sent non-200 response. Please check MLP url and input.")
        processed_text = response.json()
        return processed_text

    @check_mlp_connection
    def lemmatize(self, text):
        if not text:
            return text
        payload = {"text": text.encode(), "analyzers": ["lemmas"]}
        response = requests.post(self.mlp_url, data=payload)
        if response.status_code != 200:
            raise exceptions.MLPFailedError("MLP service sent non-200 response. Please check MLP url and input.")
        lemmas = response.json()["text"]["lemmas"]
        return lemmas


def get_mlp_analyzer(mlp_host=None, mlp_major_version=2):
    """
    Function for retrieving an MLP object based on MLP hostname and major version (2/3).
    Also possible to use MLP from package.
    :param: str mlp_host: Hostname for the MLP service.
    :param: int mlp_major_version: Major version number (2.x.x vs 3.x.x) of MLP.
    :param: bool use_api: 
    :return: analyzer object.
    """
    if mlp_host:
        # get analyzer over api
        if mlp_major_version == 2:
            analyzer = MLP2Analyzer(MLP_HOST=mlp_host)
        else:
            analyzer = MLP3Analyzer(MLP_HOST=mlp_host)
    else:
        # get analyzer from package
        # import MLP package here so it won't throw importerror if API used and package not installed
        from texta_mlp.mlp import MLP
        analyzer = MLP()
    return analyzer
